from .money import Money
from .exchange import ExchangeRate

import time


class Account:
    """
    The Account class stores and tracks Money.
    """

    def __init__(obj, name="Account", balance=Money(0), record=None):
        """
        Create an Account object.
        """
        obj.name = name
        if type(balance) in [int, float]:  # balance -> amount
            balance = Money(balance)
        elif type(balance) == str:  # balance -> currency
            balance = Money(0, balance)
        obj.balance = balance
        if record == None:  # record is of type Record()
            record = Record()
        obj.record = record

    def __repr__(obj):
        """
        Represent an Account.
        """
        return obj.name + ": " + str(obj.balance)

    # operations
    def __add__(obj, other):
        """
        Add Money to an Account.
        """
        acc = Account(obj.name, -obj.balance, obj.record.copy())
        if type(other) == type(obj):  # account + account
            acc.balance += other.balance
            acc.record.append(Transaction(other.balance))
        else:  # account + money, account + int
            acc.balance += other
            acc.record.append(Transaction(other))
        return acc

    def __sub__(obj, other):
        """
        Subtract Money from an Account.
        """
        return obj + (-other)

    def __neg__(obj):
        """
        Negate an accounts Balance.
        """
        acc = Account(obj.name, -obj.balance, obj.record.copy())
        acc.record.append(Transaction(-obj.balance * 2, "Negate balance"))
        return acc

    def __mul__(obj, other):
        """
        Multiply an Accounts balance by a number.
        """
        acc = Account(obj.name, obj.balance, obj.record.copy())
        acc.balance *= other
        acc.record.append(
            Transaction(acc.balance - obj.balance, "Multiply by " + str(other))
        )
        return acc

    def __truediv__(obj, other):
        """
        Divide an Accounts balance by a number.
        """
        return obj * (1 / other)


class Record:
    """
    A record stores transactions.
    """

    def __init__(obj, transactions=[]):
        """
        Create a record.
        """
        obj.transactions = transactions

    def __len__(obj):
        """
        Get the length of the record.
        """
        return len(obj.transactions)

    def copy(obj):
        """
        Copy the record.
        """
        return Record(obj.transactions.copy())

    def __repr__(obj):
        """
        Display the stored transactions.
        """
        recordRepresentation = "----- Record -----\n"
        for transaction in obj.transactions:
            recordRepresentation += str(transaction) + "\n"
        if len(obj) == 0:
            recordRepresentation += "No transactions.\n"
        return recordRepresentation

    def append(obj, transaction):
        """
        Append the transaction record.
        """
        obj.transactions += [transaction]

    def __add__(obj, other):
        """
        Append the transaction record (addition operator).
        """
        obj.append(other)


class Transaction:
    """
    Transactions are records of a money transfer at some time.
    """

    def __init__(obj, valueChange, description=None, timeStamp=None):
        """
        Create a Transaction object.
        """
        obj.valueChange = valueChange
        if description:
            obj.description = description
        if timeStamp == None:
            timeStamp = int(time.time())
        obj.time = timeStamp

    def __repr__(obj):
        """
        Represent a Transaction.
        """
        return "[" + str(obj.time) + "] Transaction: " + str(obj.valueChange)



from .money import Money


class ExchangeRate:
    """
    ExchangeRates convert Money of different currencies.
    """

    def __init__(obj, base, other, rate):  # 1 * base = rate * other
        """
        Create an ExchangeRate object.
        """
        obj.base = base
        obj.other = other
        obj.rate = rate

    def __repr__(obj):
        """
        A common representation of an ExchangeRate.
        """
        return str(Money(1, obj.base)) + " => " + str(Money(obj.rate, obj.other))

    # transform the ExchangeRate

    def inverse(obj):
        """
        Invert the ExchangeRate (A=2B => B=1/2A).
        """
        return ExchangeRate(obj.other, obj.base, 1 / obj.rate)

    # checks and getters

    def getCurrencies(obj):
        return {obj.base, obj.other}

    def getOther(obj, base):
        if obj.base == base:
            return obj.other
        elif obj.other == base:
            return obj.base

    def canConvert(obj, base, other=""):
        """
        Check if the ExchangeRate can convert between two currencies.
        """
        if base == other and base in obj.getCurrencies():
            return True

        if other == "":
            return base in obj.getCurrencies()

        return obj.getCurrencies() == {base, other}

    # transform some Money

    def convert(obj, money):
        """
        Convert money from one currency to another.
        """
        if type(money) != Money:  # only money can be converted
            raise TypeError("Can't convert " + str(type(money)) + " to money.")

        if obj.base == money.currency:  # base -> other
            return Money(money.amount * obj.rate, obj.other)
        elif obj.other == money.currency:  # base <- other
            return Money(money.amount / obj.rate, obj.base)
        else:  # unknown currency
            raise TypeError(
                "Can't convert " + money.currency + " here (" + str(obj) + ")."
            )

    # operations

    def __mul__(obj, money):  # Money() * ExchangeRate() = convertedMoney
        """
        Convert Money between currencies using the multiplication operator.
        """
        if type(money) != Money:
            raise TypeError("Can't multiply ExchangeRates by anything but money.")

        return obj.convert(money)

    def __eq__(obj, other):
        """
        Check if two ExchangeRates are equal.
        """
        if type(obj) != type(other):
            return False

        if obj.rate == other.rate:  # equality
            return obj.base == other.base and obj.other == other.other
        elif obj.rate == 1 / other.rate:  # equality of the inverse
            return obj.base == other.other and obj.other == other.base

        return False

    def __hash__(obj):
        """
        Hash an ExchangeRate.
        """
        return hash((obj.rate, obj.base, obj.other))  # TODO same for inverse


# Exchanges are collections of ExchangeRates.


class Exchange:
    """
    Exchanges store multiple ExchangeRates and convert Money using them.
    """

    def __init__(obj, *exchangeRates):
        """
        Create an Exchange object.
        """
        obj.exchangeRates = []
        for exchangeRate in exchangeRates:
            if type(exchangeRate) == ExchangeRate:
                obj.insertExchangeRate(exchangeRate)
            elif type(exchangeRate) in [list, tuple]:
                obj.insertExchangeRate(*exchangeRate)
            else:
                raise TypeError("Unknow type for an exchangeRate.")

    def __repr__(obj):
        """
        Represent an Exchange.
        """
        if len(obj):
            return "{" + "; ".join([str(e) for e in obj.exchangeRates]) + "}"
        else:
            return "{ Empty Exchange }"

    def __len__(obj):
        """
        Get the stored amount of ExchangeRates.
        """
        return len(obj.exchangeRates)

    # checks

    def canConvert(obj, base, other=""):
        """
        Check for ways to convert money from one currency to another.
        """
        for exchangeRate in obj.exchangeRates:
            if exchangeRate.canConvert(base, other):
                return True

        if other == "":
            return False

        for exchangeRate1 in obj.exchangeRates:  # CB
            for exchangeRate2 in obj.exchangeRates:  # AB
                if not exchangeRate1.canConvert(base, exchangeRate2.base):  # CB C, B
                    continue
                if not exchangeRate2.canConvert(exchangeRate2.base, other):  # AB A, A
                    continue
                return True

        return False

    # modify Exchanges

    def insertExchangeRate(obj, *args):  # update / append a rate
        """
        Append the Exchange by another ExchangeRate (tuple, list or obj).
        """
        if len(args) == 1:  # rate
            newExchangeRate = args[0]
        elif len(args) == 3:  # base, other, rate
            newExchangeRate = ExchangeRate(args[0], args[1], args[2])
        else:
            raise ValueError("Invalid exchange rate for Exchange.insertExchangeRate().")

        existingRates = [
            exchangeRate
            for exchangeRate in obj.exchangeRates
            if newExchangeRate.getCurrencies() == exchangeRate.getCurrencies()
        ]

        if existingRates == []:
            obj.exchangeRates += [newExchangeRate]
        else:
            oldExchangeRate = existingRates[0]
            if oldExchangeRate.base == newExchangeRate.base:
                oldExchangeRate.rate = newExchangeRate.rate
            else:
                oldExchangeRate.rate = 1 / newExchangeRate.rate

    # transform Money - unlimited steps of conversion possible :) - @finnmglas
    def exchangeRatePath(
        obj, currency, other, forwardPath=[], backwardPath=[], verbose=False
    ):
        """
        Find a path of ExchangeRates to convert one currency to another.
        """
        if verbose:
            print("Converters:", currency, "=>", other)
            print("\t", forwardPath, backwardPath)

        if currency == other:  # for 1, 2, 4, 6 conversions
            return forwardPath + backwardPath

        forwardOptions = [
            exchangeRate
            for exchangeRate in obj.exchangeRates
            if currency in exchangeRate.getCurrencies()
            and exchangeRate not in forwardPath
        ]
        backwardOptions = [
            exchangeRate
            for exchangeRate in obj.exchangeRates
            if other in exchangeRate.getCurrencies()
            and exchangeRate not in backwardPath
        ]

        if not len(forwardOptions) or not len(backwardOptions):
            raise ValueError(
                "Can't convert the currencies "
                + currency
                + " and "
                + other
                + " using this exchange."
            )

        if verbose:
            print(currency, "can be converted using", forwardOptions)
            print(other, "can be converted using", backwardOptions)

        common = [
            exchangeRate
            for exchangeRate in forwardOptions
            if exchangeRate in backwardOptions
        ]

        if common != []:  # for 3, 5, 7 conversions
            if verbose:
                print("Common conversions:", common)
            return forwardPath + [common[0]] + backwardPath

        if verbose:
            print("No Solution yet")

        possibleContinuations = [
            (forwardOption, backwardOption)
            for forwardOption in forwardOptions
            for backwardOption in backwardOptions
        ]
        for forwardOption, backwardOption in possibleContinuations:
            common = obj.exchangeRatePath(
                forwardOption.getOther(currency),
                backwardOption.getOther(other),
                forwardPath + [forwardOption],
                [backwardOption] + backwardPath,
                verbose,
            )
            return common

    def convert(obj, money, destinationCurrency, verbose=False):
        """
        Convert Money to a destination-currency.
        """
        if type(money) != Money:
            raise TypeError("Can't convert " + str(type(money)) + " to money.")

        if money.currency == destinationCurrency:
            return money
        # get conversions path
        conversions = obj.exchangeRatePath(
            money.currency, destinationCurrency, verbose=verbose
        )
        if verbose:
            print("ExchangePath:", conversions)
            print("Money:", money)

        convertedMoney = money
        # 'walk' the conversions path
        for exchangeRate in conversions:
            convertedMoney = exchangeRate.convert(convertedMoney)
            if verbose:
                print("Converted to:", convertedMoney)

        return convertedMoney



from .fetch import ExchangeRateFetcher


class Money:
    """
    Money has an amount and a currency. It is a core type to be acted upon.
    """

    defaultCurrency = "EUR"
    exchange = None

    # --- Convert currencies

    @staticmethod
    def ensureExchangeExists():
        """
        The Money class has an associated Exchange. Ensure it exists.
        """
        if Money.exchange == None:
            from .exchange import Exchange

            Money.exchange = Exchange()

    @staticmethod
    def addExchangeRate(*exchangeRate):
        """
        Add an ExchangeRate to the global money Exchange.
        """
        Money.ensureExchangeExists()
        Money.exchange.insertExchangeRate(*exchangeRate)

    @staticmethod
    def canConvert(base, other):
        """
        Check if the money Exchange can convert from a base currency to another.
        """
        Money.ensureExchangeExists()
        return Money.exchange.canConvert(base, other)

    @staticmethod
    def fetchRates(source="ecb", verbose=False):
        """
        Fetch ExchangeRates from a source api.
        """
        for e in ExchangeRateFetcher.fetch(source, verbose):
            Money.addExchangeRate(*e)

    def to(obj, currency):
        """
        Convert the Money object using the global money Exchange to another currency.
        """
        Money.ensureExchangeExists()
        return Money.exchange.convert(obj, currency)

    # --- Constructor

    def __init__(obj, amount=0, currency=defaultCurrency):
        """
        Create a Money object.
        """
        obj.amount = amount
        obj.currency = currency

    # --- type conversions

    def __repr__(obj):
        """
        A common representation of Money.
        """
        return str(obj.amount) + " " + str(obj.currency)

    def __int__(obj):
        """
        Get the Money's amount.
        """
        return obj.amount

    def __truth__(obj):
        """
        True if there is an amount of Money stored.
        """
        return obj.amount > 0

    # --- comparisons

    def __eq__(obj, other):
        """
        Check for equality of Money and another Money's amount or a number.
        Also converts to the necessary currency.
        """
        if type(other) in [int, float]:
            return obj.amount == other
        elif type(other) == type(obj):
            return obj.amount == other.to(obj.currency).amount
        else:
            raise TypeError(
                "Can't check equality of "
                + str(type(obj))
                + " and "
                + str(type(other))
                + "."
            )

    def __ne__(obj, other):
        return not obj == other

    def __lt__(obj, other):
        if type(other) in [int, float]:
            return obj.amount < other
        elif type(other) == type(obj):
            return obj.amount < other.to(obj.currency).amount
        else:
            raise TypeError(
                "Can't compare " + str(type(obj)) + " and " + str(type(other)) + "."
            )

    def __le__(obj, other):
        if type(other) in [int, float]:
            return obj.amount <= other
        elif type(other) == type(obj):
            return obj.amount <= other.to(obj.currency).amount
        else:
            raise TypeError(
                "Can't compare " + str(type(obj)) + " and " + str(type(other)) + "."
            )

    def __gt__(obj, other):
        return not obj <= other

    def __ge__(obj, other):
        return not obj < other

    # --- calculations

    def __add__(obj, other):
        """
        Add Money objects (auto-conversion) or numbers to money.
        """
        if type(other) in [int, float]:
            return Money(obj.amount + other, obj.currency)
        elif type(other) == type(obj):
            return Money(obj.amount + other.to(obj.currency).amount, obj.currency)
        else:
            raise TypeError(
                "Can't add " + str(type(obj)) + " and " + str(type(other)) + "."
            )

    def __sub__(obj, other):
        """
        Subract Money objects (auto-conversion) or numbers from money.
        """
        return obj + (-other)

    def __neg__(obj):
        """
        Negate the Money's amount.
        """
        return Money(-obj.amount, obj.currency)

    def __mul__(obj, other):
        """
        Multiplie Money by a number (not by Money).
        When multiplied by an ExchangeRate, the Money is converted by it.
        """
        from .exchange import ExchangeRate

        if type(other) == ExchangeRate:
            return other * obj  # conversion
        elif type(other) in [int, float]:
            return Money(obj.amount * other, obj.currency)
        else:
            raise TypeError(
                "Can't multiply " + str(type(obj)) + " by " + str(type(other)) + "."
            )

    def __truediv__(obj, other):
        """
        Divide Money by a number (=> Money) or Money (=> Number).
        """
        if type(other) in [int, float]:
            return Money(obj.amount / other, obj.currency)
        elif type(other) == type(obj):
            return obj.amount / other.to(obj.currency).amount
        else:
            raise TypeError(
                "Can't divide " + str(type(obj)) + " by " + str(type(other)) + "."
            )

    def __mod__(obj, other):
        """
        Return the leftover number when dividing money by an integer.
        """
        if type(other) in [int, float]:
            return Money(obj.amount % other, obj.currency)
        elif type(other) == type(obj):
            return obj.amount % other.to(obj.currency).amount
        else:
            raise TypeError(
                "Can't mod " + str(type(obj)) + " by " + str(type(other)) + "."
            )



import requests
from xml.etree import ElementTree as XMLElementTree


class ExchangeRateFetcher:
    """
    An API connector for ledgerman that fetches ExchangeRates.
    """

    @staticmethod
    def printExchangeRate(e):
        print(e[0], "=>", e[2], e[1])

    @staticmethod
    def fetch(source="ecb", verbose=False):
        """
        Fetch ExchangeRates from a source.
        """
        source = source.lower()
        exchangeRates = []
        # fetch, convert and return
        if source == "ecb":
            exchangeRates = ExchangeRateFetcher.fetch_ecb(verbose)
        elif source == "exchangeratesapi_io":
            exchangeRates = ExchangeRateFetcher.fetch_exchangeratesapi_io(verbose)
        elif source == "coingecko":
            exchangeRates = ExchangeRateFetcher.fetch_coingecko(verbose)
        else:
            raise ValueError("ExchangeRateFetcher: Invalid Source: " + source)
        return exchangeRates

    # APIs to fetch from
    @staticmethod
    def fetch_ecb(verbose=False):  # REAL: official, XML, Free
        """
        Fetch ExchangeRates from the european central bank (updated daily).
        """
        exchangeRates = []
        # fetch
        r = requests.get(
            "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml"
        )
        if r.status_code != 200:
            raise IOError("Connecting to the ECB XML API failed.")
        # parse xml
        tree = XMLElementTree.fromstring(r.text)
        date = tree[2][0].attrib["time"]
        ratesData = tree[2][0]
        # convert data to ExchangeRates
        if verbose:
            print("ExchangeRates (ECB, " + date + "):")
        for d in ratesData:
            e = "EUR", d.attrib["currency"], float(d.attrib["rate"])
            if verbose:
                ExchangeRateFetcher.printExchangeRate(e)
            exchangeRates += [e]
        return exchangeRates

    @staticmethod
    def fetch_exchangeratesapi_io(verbose=False):  # REAL: inofficial, JSON, Free
        """
        Fetch ExchangeRates from api.exchangeratesapi.io (which takes input from ecb).
        """
        exchangeRates = []
        # fetch
        r = requests.get("https://api.exchangeratesapi.io/latest")
        if r.status_code != 200:
            raise IOError("Connecting to the exchangeratesapi.io JSON API failed.")
        # parse json
        jsonResult = r.json()
        rates = jsonResult["rates"]
        base = jsonResult["base"]
        date = jsonResult["date"]
        # convert data to ExchangeRates
        if verbose:
            print("ExchangeRates (exchangeratesapi.io, " + date + "):")
        for currency in rates.keys():
            e = base, currency, rates[currency]
            if verbose:
                ExchangeRateFetcher.printExchangeRate(e)
            exchangeRates += [e]
        return exchangeRates

    @staticmethod
    def fetch_coingecko(verbose=False):  # CRYPTO: inofficial, JSON, Free
        """
        Fetch ExchangeRates from api.coingecko.com.
        Docs: https://www.coingecko.com/api/documentations/v3
        """
        exchangeRates = []

        coingecko_sym2id = {  # see https://api.coingecko.com/api/v3/coins/list
            "eth": "ethereum",
            "btc": "bitcoin",
            "ltc": "litecoin",
            "trx": "tron",
            "usdt": "tether",
            "xrp": "ripple",
            "link": "chainlink",
            "best": "bitpanda-ecosystem-token",
            "pan": "pantos",
            "miota": "iota",
            "ada": "cardano",
            "vet": "vechain",
            "omg": "omisego",
            "neo": "neo",
            "qtum": "qtum",
            "xem": "nem",
            "xtz": "tezos",
            "yfi": "yearn-finance",
            "chz": "chiliz",
            "xlm": "stellar",
            "ont": "ontology",
            "bch": "bitcoin-cash",
            "usdc": "usd-coin",
            "eos": "eos",
            "uni": "uniswap",
            "waves": "waves",
            "atom": "cosmos",
            "dot": "polkadot",
            "snx": "havven",
            "dash": "dash",
            "zrx": "0x",
            "bat": "basic-attention-token",
            "kmd": "komodo",
            "etc": "ethereum-classic",
            "doge": "dogecoin",
            "zec": "zcash",
            "rep": "augur",
            "lsk": "lisk",
            "comp": "compound-governance-token",
            "mkr": "maker",
        }
        coingecko_id2sym = {v: k for k, v in coingecko_sym2id.items()}
        # build query
        request_url = "https://api.coingecko.com/api/v3/simple/price?ids="
        request_url += ",".join(coingecko_sym2id.values())
        request_url += "&vs_currencies=eur"
        # fetch
        r = requests.get(request_url)
        if r.status_code != 200:
            raise IOError("Connecting to the coingecko.com JSON API failed.")
        # parse json
        jsonResult = r.json()

        # convert data to ExchangeRates
        if verbose:
            print("ExchangeRates (coingecko.com):")
        for id in jsonResult.keys():
            rates = jsonResult[id]
            e = coingecko_id2sym[id].upper(), "EUR", rates["eur"]
            if verbose:
                ExchangeRateFetcher.printExchangeRate(e)
            exchangeRates += [e]
        return exchangeRates



from setuptools import setup

long_description = """
|PyPI - Version| |Downloads| |Discord| |Code style: black|

LedgerMan
=========

   Yet another python library for finance. **Why?**

-  LedgerMan is **comprehensive**. Check out `its usage`_!
-  LedgerMan is **open, welcoming and transparent**: `join us on
   discord`_!
-  LedgerMan provides **powerful** financial tools and models.

=====

`View on GitHub`_, `contact Finn`_ or `sponsor this project ❤️`_!

.. _its usage: https://github.com/finnmglas/ledgerman#usage
.. _join us on discord: https://discord.com/invite/BsZXaur
.. _View on GitHub: https://github.com/finnmglas/ledgerman
.. _contact Finn: https://contact.finnmglas.com
.. _sponsor this project ❤️: https://sponsor.finnmglas.com

.. |PyPI - Version| image:: https://img.shields.io/pypi/v/ledgerman?color=000
   :target: https://pypi.org/project/ledgerman/
.. |Downloads| image:: https://img.shields.io/badge/dynamic/json?style=flat&color=000&maxAge=10800&label=downloads&query=%24.total_downloads&url=https%3A%2F%2Fapi.pepy.tech%2Fapi%2Fprojects%2Fledgerman
   :target: https://pepy.tech/project/ledgerman
.. |Discord| image:: https://img.shields.io/badge/discord-join%20chat-000
   :target: https://discord.com/invite/BsZXaur
.. |Code style: black| image:: https://img.shields.io/badge/code%20style-black-000000.svg
   :target: https://github.com/psf/black
"""

setup(
    name="ledgerman",
    version="0.5.1",
    description="The python library for finance.",
    long_description=long_description,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Office/Business :: Financial :: Accounting",
    ],
    keywords="accounting finance manager money library ledger ledgerman crypto",
    url="http://github.com/finnmglas/LedgerMan",
    author="Finn M Glas",
    author_email="finn@finnmglas.com",
    license="MIT",
    packages=["ledgerman"],
    entry_points={
        "console_scripts": [
            "pymoney=ledgerman.tools:PyMoney.main",
            "ledgerman=ledgerman.tools:LedgerMan.main",
        ],
    },
    install_requires=["requests", "argparse"],
    test_suite="nose.collector",
    tests_require=["nose"],
    include_package_data=True,
    zip_safe=False,
)
